/******************************************************************************
 * The virtual class DB for All the DB Related Operations. This is indenpent
 * of the database platform.
 * Author: Yuanyuan Tian (ytian@umich.edu)
 * ***************************************************************************/

#ifndef DB_H
#define DB_H

#include <string>
#include <vector>
#include <hash_map.h>
#include <map>
#include "../util/util.h"

using namespace std;

/*
 * BASE TABLES:
 *   graph(graphID, attrName, attrType, attrValue);
 *   node(graphID, nodeID, attrName, attrType, attrValue);
 *   edge(graphID, node1ID, node2ID, attrName, attrType, attrValue);
 *   ortholog(orthologID, orthologName, source);
 * 
 * INDEXES FOR BASE TABLES: (OPTIONAL)
 *   graph_index on graph(graphID, attrName, attrType, attrValue);
 *   node_index on node(graphID, nodeID, attrName, attrType, attrValue);
 *   edge_index on edge(graphID, node1ID, node2ID, attrName, attrType, attrValue);
 *   ortholog_index on ortholog(orthlogID);
 * 
 * TABLES FOR SPECIAL INDEXES:
 *   distance(graphID, node1ID, node2ID, distance);
 *   fragment(orthSeq, distSum, graphID, nidSeq, distMatrix);
 *   neighborhood(orthlogID, degree, nbConnection, nbArray, graphID, nodeID);
 * 
 * SPECIAL INDEXES: (NECESSARY)
 *   distance_index on distance (graphID, node1ID, node2ID, distance);
 *   fragment_index on fragment(orthSeq, distSum);
 *   neighborhood_index on neighborhood(orthologID, degree, nbConnection); 
 * 
 * ALLOWED ATTRIBUTE TYPES: 
 *   int: i, 
 *   float: f, 
 *   char: c, 
 *   string: s, 
 *   bool: b 
 */
/*
enum AttributeType
{
	INT, FLOAT, CHAR, STRING, BOOL
};*/

typedef char AttributeType;

struct AttributeValue
{
	AttributeType type;
	string value;
};

typedef hash_multimap<string, AttributeValue, strhash, eqstr> Attributes;
//typedef multimap<string, AttributeValue> Attributes;

class DB
{
public:
//	virtual DB()=0;
//virtual	~DB(){};
	//-------------------------------------------------------------------------
	// Operations on Graph Table
	//-------------------------------------------------------------------------
	
	//create the graph table 
	//schema: graph(graphID, attrName, attrType, attrValue)
	//input: the table name
	virtual void createGraphTable(string tablename)=0;
	
	//insert one graph into the graph table
	virtual void insertIntoGraphTable(
			string tablename, 
			int graphID, 
			const Attributes& attrs //the attrbutes of the graph
			)=0;
	
	//create an index for getting graph attribute(s)
	//graph_index(graphID, attrName, attrType, attrValue) on graph
	virtual void createGraphIndex(string tablename)=0;
	
	//access one attribute of a graph
	virtual void getGraphAttr(
			string tablename, 
			int graphID, 
			string attrName,
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//access all the attributes of a graph
	virtual void getGraphAttr(
			string tablename, // input
			int graphID, //input
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//get the maximun graphID in the graph table
	virtual int getMaxGraphID(string tablename)=0;
	
	//select graphs based on the given condition
	virtual void selectGraphs(
			string tablename, //input
			string condition, //the input sql condition on the graph table
							  //e.g. "where attrName='name'"
			vector<int>& graphs //return value
			)=0;
	
	//-------------------------------------------------------------------------
	// Opertaions on Node Table
	//-------------------------------------------------------------------------
	
	//create the node table
	//schema: node(graphID, nodeID, attrName, attrType, attrValue)
	//input: the table name
	virtual void createNodeTable(string tablename)=0;
	
	//insert one node into the node table
	//node_index(graphID, nodeID, attrName, attrType, attrValue)
	virtual void insertIntoNodeTable(
			string tablename, 
			int graphID, 
			int nodeID, 
			const Attributes& attrs //the attributes of the node
			)=0;
	
	//create an index for getting node attribute(s)
	virtual void createNodeIndex(string tablename)=0;
	
	//access one attribute of a node
	virtual void getNodeAttr(
			string tablename, 
			int graphID, 
			int nodeID, 
			string attrName,
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//access all the attributes of a node
	virtual void getNodeAttr(
			string tablename, //input
			int graphID, //input
			int nodeID, //input
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//select nodes based on the given condition
	virtual void selectNodes(
			string tablename, //input
			string condition, //the input sql condition on the node table
			vector<int>& graphs,//return value
			vector<int>& nodes, //return value
			vector<Attributes*>& vattrs	//return value
			)=0;
	
		//group nodes by attribute values, for categorical attribute
	virtual void groupNodesByAttr(
			string tablename, //input
			int graphID, //input
			string attrName, //input
			vector<int>& nodes, //return value
			vector<string>& attrValues //return value
			)=0;

	//group nodes by attribute values,for numerical attribute
	virtual void groupNodesByAttrnumeric(
			string tablename, //input
			int graphID, //input
			string attrName, //input
			vector<int>& nodes, //return value
			vector<int>& attrValues //return value
			)=0;
	
	
	//-------------------------------------------------------------------------
	// Opertaions on Edge Table
	//-------------------------------------------------------------------------
	
	//create the edge table
	//schema: edge(graphID, node1ID, node2ID, attrName, attrType, attrValue)
	//input: table name
	virtual void createEdgeTable(string tablename)=0;
	
	//insert one edge into the edge table
	virtual void insertIntoEdgeTable(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			const Attributes& attrs //the attributes of the edge
			)=0;

	//create an index for getting edge attribute(s)
	//edge_index(graphID, node1ID, node2ID, attrName, attrType, attrValue)
	virtual void createEdgeIndex(string tablename)=0;
	
	//access one attribute of an edge
	virtual void getEdgeAttr(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			string attrName,
			string& attrs //return value (passed by reference)
			)=0;

	//access one attribute of an edge
	virtual void getEdgeAttr(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			string attrName,
			Attributes& attrs //return value (passed by reference)
			)=0;
	
	//access all the attributes of an edge
	virtual void getEdgeAttr(
			string tablename, //input
			int graphID, //input
			int node1ID, //input
			int node2ID, //input
			Attributes& attrs //return value
			)=0;
	
	//select nodes based on the given condition
	virtual void selectEdges(
			string tablename, //input
			string condition, //the input sql condition on the edge table
			vector<int>& graphs, //return value
			vector<int>& node1s, //return value
			vector<int>& node2s, //return
			vector<Attributes*>& vattrs	//return value
			)=0;
	
	//select the imediate neighbors of a node based on the given condition
	virtual void selectNeighbors(
			string tablename, //input
			int graphID, //input
			int nodeID, //input
			string condition, //the input sql condition on the edge table
			vector<int>& nodes, //return value
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	
	//-------------------------------------------------------------------------
	// Opertaions on Ortholog Table
	//-------------------------------------------------------------------------
	
	//create the ortholog table
	//schema: ortholog(orthologID, orthologName, source)
	//input: table name
	virtual void createOrthologTable(string tablename)=0;
	
	//insert one ortholog into the ortholog table
	virtual void insertIntoOrthologTable(
			string tablename, 
			int orthologID, 
			string orthologName, 
			string source
			)=0;
	
	
	//create an index for getting ortholog
	//ortholog_index(orthlogID) on ortholog;
	virtual void createOrthologIndex(string tablename)=0;
	
	//select the orthologName and source given a orthologID
	virtual void getOrtholog(
			string tablename, //input
			int orthologID, //input
			string& orthologName, //return value 
			string& source, //return value
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//-------------------------------------------------------------------------
	// Operations on Distance Table
	//-------------------------------------------------------------------------
	
	//create distance table
	//distance(graphID, node1ID, node2ID, distance)
	virtual void createDistanceTable(string tablename)=0;

	//insert one tuple into the distance table
	virtual void insertIntoDistanceTable(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			int distance
			)=0;
	
	//create the distance index
	//distance_index(graphID, node1ID, node2ID, distance) on distance
	virtual void createDistanceIndex(string tablename)=0;
	
	//create the clustered distance index (distance table is sorted)
	//distance_index(graphID, node1ID, node2ID, distance) on distance
//	virtual void createClusteredDistanceIndex(string tablename)=0;

	//select the distance from node1 to node2 in the graph
	virtual int selectDistance(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID,
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//select the nearby nodes with a certain distance with the node
	virtual void selectNearbyNodes(
			string tablename, //input
			int graphID, //input
			int nodeID, //input
			int distance, //input
			vector<int>& nodes, //return value
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	//-------------------------------------------------------------------------
	// Operations on Fragment Table
	//-------------------------------------------------------------------------
	
	//create the fragment table
	//fragment(orthSeq, distSum, graphID, nidSeq, distMatrix)
	virtual void createFragmentTable(string tablename)=0;
	
	//insert one tuple into the fragment table
	virtual void insertIntoFragmentTable(
			string tablename, 
			int graphID, 
			const vector<int>& nodes,
			const vector<uint32_t>& distMatrix, 
			int summary, 
			const vector<int>& orthlogs
			)=0;
	
	
	//create the fragment index
	//fragment_index(orthSeq, distSum) on fragment
	virtual void createFragmentIndex(string tablename)=0;
	
	//create the clustered fragment index (fragment table is sorted)
	//fragment_index(orthSeq, distSum) on fragment
	virtual void createClusteredFragmentIndex(string tablename)=0;
	
	//select fragments given the query fragment and the allowed the difference
	virtual void selectFragments(
			string tablename, //input
			const uint64_t orthSeq, //input
			int distSumLower, //input
			int distSumUpper, //input
			vector<int>& graphs, //return value, the graphIDs of 
								//matching fragment
			vector<vector<int> >& fragments, //return value, the nodeIDs of 
											//the matching fragments
			vector<vector<int> >& matrices, //
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			)=0;
	
	
	//-------------------------------------------------------------------------
	// Operations on Neighborhood Table
	//-------------------------------------------------------------------------
	
	//create the neighborhood table
	//neighborhood(orthlogID, degree, nbConnection, nbArray, graphID, nodeID)
	virtual void createNeighborhoodTable(string tablename)=0;
	
	//insert one tuple into the neighborhood table
	virtual void insertIntoNeighborhoodTable(
			string tablename, 
			int graphID, 
			int nodeID, 
			int orthologID, 
			int degree, 
			int nbConnection, 
			const vector<bool>& nbArray
			)=0;
	
	//create the neighborhood index
	//neighborhood_index(orthologID, degree, nbConnection) on neighborhood
	virtual void createNeighborhoodIndex(string tablename)=0;
	
	//create the clustered neighborhood index (neighborhood table is sorted)
	//neighborhood_index(orthologID, degree, nbConnection) on neighborhood	
	virtual void createClusteredNeighborhoodIndex(string tablename)=0;
	
	//select the nodes that match the query neighborhood given the
	//allowed the differences in number of matching neighbors and neighbor
	//connections
	virtual void selectNeighborhoods(
			string tablename, //input
			vector<int>& orths, //input
			int degree, //input
			int neighborDiff, //input
			int nbConnection, //input
			int connDiff, //input
			const bool* nbArray, //input
			vector<int>& graphs, //return value, the graphIDs of the 
								 //matching nodes 
			vector<int>& nodes, //return value, the nodeIDs of the
							   //matching nodes
			vector<float>& scores //return value, the score of each
								//mathing node
			)=0;
	
	//-------------------------------------------------------------------------
	// Copy Operations
	//-------------------------------------------------------------------------
	
	//copy a table to a file using a specifed copy mode (e.g binary or text)
	virtual void copyTableTo(
			string tablename, 
			string mode, 
			string path //the "absolute" path of the file
			)=0;
	
	//copy a file to a table using a specifed copy mode (e.g binary or text)
	virtual void copyTableFrom(
			string tablename, 
			string mode, 
			string path //the "absolute" path of the file
			)=0;
	
	//-------------------------------------------------------------------------
	// Prepare Operation
	//-------------------------------------------------------------------------
	
	//prepare query plans, to avoid the overhead of finding query plans at
	//query execution times
	virtual void preparePlan(string plan)=0;
};
#endif
