/******************************************************************************
 * The class PGDB which inherits the DB. It is implemented using PostgreSQL
 * Author: Yuanyuan Tian (ytian@umich.edu)
 * ***************************************************************************/
#ifndef PGDB_H
#define PGDB_H

#include <sys/types.h>
#include "libpq-fe.h"
#include <stdio.h>
#include <stdlib.h>
#include "db.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include "../util/util.h"

/*
 * BASE TABLES:
 *   create table graph(
 * 						graphID int4, 
 * 						attrName varchar, 
 * 						attrType char(1), 
 * 						attrValue varchar
 * 						);
 *   create table node(
 * 						graphID int4, 
 * 						nodeID int4, //note that for SAGA, the nodeID<2^16
 * 						attrName varchar, 
 * 						attrType char(1), 
 * 						attrValue varchar
 * 						);
 *   create table edge(
 * 						graphID int4, 
 * 						node1ID int4, 
 * 						node2ID int4, 
 * 						attrName varchar, 
 * 						attrType char(1), 
 * 						attrValue varchar
 * 						);
 *   create table ortholog(
 * 						orthologID int4, 
 * 						orthologName varchar, 
 * 						source varchar
 * 						);
 * 
 * INDEXES FOR BASE TABLES: (OPTIONAL)
 *   create index graph_index 
 * 			on graph(graphID, attrName, attrType, attrValue);
 *   create index node_index 
 * 			on node(graphID, nodeID, attrName, attrType, attrValue);
 *   create index edge_index 
 * 			on edge(graphID, node1ID, node2ID, attrName, attrType, attrValue);
 *   create index edge_index2 
 * 			on edge(graphID, node2ID, node1ID);
 *   create index ortholog_index 
 * 			on ortholog(orthlogID);
 * 
 * TABLES FOR SPECIAL INDEXES:
 *   create table distance(
 * 						graphID int4, 
 * 						node1ID int4, 
 * 						node2ID int4, 
 * 						distance int4
 * 						);
 *   create table fragment(
 * 						orthSeq int8, 
 * 						distSum int4, 
 * 						graphID int4, 
 * 						nidSeq int4[], 
 * 						distMatrix int4[]
 * 						);
 *   create table neighborhood(
 * 						orthlogyID int4 , 
 * 						degree int4, 
 * 						nbConnection int4, 
 * 						nbArray bit(96), 
 * 						graphID int4, 
 * 						nodeID int4
 * 						);
 * 
 * SPECIAL INDEXES: (NECESSARY)
 *   create index distance_index 
 * 			on distance (graphID, node1ID, node2ID, distance);
 *   create index fragment_index 
 * 			on fragment(orthSeq, distSum);
 *   create index neighborhood_index 
 * 			on neighborhood(orthologID, degree, nbConnection);
 * 
 * attrType:
 *   i: int
 *   f: float
 *   c: char
 *   s: string
 *   b: bool
 */

class PGDB: public DB
{
public:
	//constructor
	// inputs: database name, user name, and user password
	PGDB(char* dbname, char* username, char* password)
	{
		char conninfo[100];
		sprintf(conninfo, "dbname=%s user=%s password=%s", 
				dbname, username, password);
		connectDB(conninfo);
	}
	
	//destructor
	~PGDB()
	{
		disconnectDB();
	}
	//create the graph table 
	//schema: graph(graphID, attrName, attrType, attrValue)
	//input: the table name
	void createGraphTable(string tablename);
	
	//insert one graph into the graph table
	void insertIntoGraphTable(
			string tablename, 
			int graphID, 
			const Attributes& attrs //the attrbutes of the graph
			);
	
	//create an index for getting graph attribute(s)
	//graph_index(graphID, attrName, attrType, attrValue) on graph
	void createGraphIndex(string tablename);
	
	//access one attribute of a graph
	void getGraphAttr(
			string tablename, 
			int graphID, 
			string attrName,
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//access all the attributes of a graph
	void getGraphAttr(
			string tablename, // input
			int graphID, //input
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//get the maximun graphID in the graph table
	int getMaxGraphID(string tablename);
	
	//select graphs based on the given condition
	void selectGraphs(
			string tablename, //input
			string condition, //the input sql condition on the graph table
							  //e.g. "where attrName='name'"
			vector<int>& graphs //return value
	);
	
	//-------------------------------------------------------------------------
	// Opertaions on Node Table
	//-------------------------------------------------------------------------
	
	//create the node table
	//schema: node(graphID, nodeID, attrName, attrType, attrValue)
	//input: the table name
	void createNodeTable(string tablename);
	
	//insert one node into the node table
	//node_index(graphID, nodeID, attrName, attrType, attrValue)
	void insertIntoNodeTable(
			string tablename, 
			int graphID, 
			int nodeID, 
			const Attributes& attrs //the attributes of the node
			);
	
	//create an index for getting node attribute(s)
	void createNodeIndex(string tablename);
	
	//access one attribute of a node
	void getNodeAttr(
			string tablename, 
			int graphID, 
			int nodeID, 
			string attrName,
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//access all the attributes of a node
	void getNodeAttr(
			string tablename, //input
			int graphID, //input
			int nodeID, //input
			Attributes& attrs, //return value (passed by reference)
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//select nodes based on the given condition
	void selectNodes(
			string tablename, //input
			string condition, //the input sql condition on the node table
			vector<int>& graphs,//return value
			vector<int>& nodes, //return value
			vector<Attributes*>& vattrs	//return value
			);
	
	//group nodes by attribute values,for categorical attribute
	void groupNodesByAttr(
			string tablename, //input
			int graphID, //input
			string attrName, //input
			vector<int>& nodes, //return value
			vector<string>& attrValues //return value
			);

	//group nodes by attribute values,for numerical attribute
	void groupNodesByAttrnumeric(
			string tablename, //input
			int graphID, //input
			string attrName, //input
			vector<int>& nodes, //return value
			vector<int>& attrValues //return value
			);
	
	//-------------------------------------------------------------------------
	// Opertaions on Edge Table
	//-------------------------------------------------------------------------
	
	//create the edge table
	//schema: edge(graphID, node1ID, node2ID, attrName, attrType, attrValue)
	//input: table name
	void createEdgeTable(string tablename);
	
	//insert one edge into the edge table
	void insertIntoEdgeTable(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			const Attributes& attrs //the attributes of the edge
			);

	//create an index for getting edge attribute(s)
	//edge_index(graphID, node1ID, node2ID, attrName, attrType, attrValue)
	void createEdgeIndex(string tablename);
	
	//access one attribute of an edge
	void getEdgeAttr(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			string attrName,
			string& attrs //return value (passed by reference)
			);
	

	//access one attribute of an edge
	void getEdgeAttr(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			string attrName,
			Attributes& attrs //return value (passed by reference)
			);
	
	//access all the attributes of an edge
	void getEdgeAttr(
			string tablename, //input
			int graphID, //input
			int node1ID, //input
			int node2ID, //input
			Attributes& attrs //return value
			);
	
	//select nodes based on the given condition
	void selectEdges(
			string tablename, //input
			string condition, //the input sql condition on the edge table
			vector<int>& graphs, //return value
			vector<int>& node1s, //return value
			vector<int>& node2s, //return
			vector<Attributes*>& vattrs	//return value
			);
	
	//select the imediate neighbors of a node based on the given condition
	void selectNeighbors(
			string tablename, //input
			int graphID, //input
			int nodeID, //input
			string condition, //the input sql condition on the edge table
			vector<int>& nodes, //return value
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	
	//-------------------------------------------------------------------------
	// Opertaions on Ortholog Table
	//-------------------------------------------------------------------------
	
	//create the ortholog table
	//schema: ortholog(orthologID, orthologName, source)
	//input: table name
	void createOrthologTable(string tablename);
	
	//insert one ortholog into the ortholog table
	void insertIntoOrthologTable(
			string tablename, 
			int orthologID, 
			string orthologName, 
			string source
			);
	
	
	//create an index for getting ortholog
	//ortholog_index(orthlogID) on ortholog;
	void createOrthologIndex(string tablename);
	
	//select the orthologName and source given a orthologID
	void getOrtholog(
			string tablename, //input
			int orthologID, //input
			string& orthologName, //return value 
			string& source, //return value
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//-------------------------------------------------------------------------
	// Operations on Distance Table
	//-------------------------------------------------------------------------
	
	//create distance table
	//distance(graphID, node1ID, node2ID, distance)
	void createDistanceTable(string tablename);

	//insert one tuple into the distance table
	void insertIntoDistanceTable(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID, 
			int distance
			);
	
	//create the distance index
	//distance_index(graphID, node1ID, node2ID, distance) on distance
	void createDistanceIndex(string tablename);
	
	//create the clustered distance index (distance table is sorted)
	//distance_index(graphID, node1ID, node2ID, distance) on distance
//	void createClusteredDistanceIndex(string tablename);

	//select the distance from node1 to node2 in the graph
	int selectDistance(
			string tablename, 
			int graphID, 
			int node1ID, 
			int node2ID,
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//select the nearby nodes with a certain distance with the node
	void selectNearbyNodes(
			string tablename, //input
			int graphID, //input
			int nodeID, //input
			int distance, //input
			vector<int>& nodes, //return value
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	//-------------------------------------------------------------------------
	// Operations on Fragment Table
	//-------------------------------------------------------------------------
	
	//create the fragment table
	//fragment(orthSeq, distSum, graphID, nidSeq, distMatrix)
	void createFragmentTable(string tablename);
	
	//insert one tuple into the fragment table
	void insertIntoFragmentTable(
			string tablename, 
			int graphID, 
			const vector<int>& nodes,
			const vector<uint32_t>& distMatrix, 
			int summary, 
			const vector<int>& orthlogs
			);
	
	
	//create the fragment index
	//fragment_index(orthSeq, distSum) on fragment
	void createFragmentIndex(string tablename);
	
	//create the clustered fragment index (fragment table is sorted)
	//fragment_index(orthSeq, distSum) on fragment
	void createClusteredFragmentIndex(string tablename);
	
	//select fragments given the query fragment and the allowed the difference
	void selectFragments(
			string tablename, //input
			const uint64_t orthSeq, //input
			int distSumLower, //input
			int distSumUpper, //input
			vector<int>& graphs, //return value, the graphIDs of 
								//matching fragment
			vector<vector<int> >& fragments, //return value, the nodeIDs of 
											//the matching fragments
			vector<vector<int> >& matrices, //
			string preparedplan //the name of a prepared plan
								//NULL if you don't want to use
								//prepared plan
			);
	
	
	//-------------------------------------------------------------------------
	// Operations on Neighborhood Table
	//-------------------------------------------------------------------------
	
	//create the neighborhood table
	//neighborhood(orthlogID, degree, nbConnection, nbArray, graphID, nodeID)
	void createNeighborhoodTable(string tablename);
	
	//insert one tuple into the neighborhood table
	void insertIntoNeighborhoodTable(
			string tablename, 
			int graphID, 
			int nodeID, 
			int orthologID, 
			int degree, 
			int nbConnection, 
			const vector<bool>& nbArray
			);
	
	//create the neighborhood index
	//neighborhood_index(orthologID, degree, nbConnection) on neighborhood
	void createNeighborhoodIndex(string tablename);
	
	//create the clustered neighborhood index (neighborhood table is sorted)
	//neighborhood_index(orthologID, degree, nbConnection) on neighborhood	
	void createClusteredNeighborhoodIndex(string tablename);
	
	//select the nodes that match the query neighborhood given the
	//allowed the differences in number of matching neighbors and neighbor
	//connections
	void selectNeighborhoods(
			string tablename, //input
			vector<int>& orths, //input
			int degree, //input
			int neighborDiff, //input
			int nbConnection, //input
			int connDiff, //input
			const bool* nbArray, //input
			vector<int>& graphs, //return value, the graphIDs of the 
								 //matching nodes 
			vector<int>& nodes, //return value, the nodeIDs of the
							   //matching nodes
			vector<float>& scores //return value, the score of each
								//mathing node
			);
	
	//-------------------------------------------------------------------------
	// Copy Operations
	//-------------------------------------------------------------------------
	
	//copy a table to a file using a specifed copy mode (e.g binary or text)
	void copyTableTo(
			string tablename, 
			string mode, 
			string path //the "absolute" path of the file
			);
	
	//copy a file to a table using a specifed copy mode (e.g binary or text)
	void copyTableFrom(
			string tablename, 
			string mode, 
			string path //the "absolute" path of the file
			);
	
	//-------------------------------------------------------------------------
	// Prepare Operation
	//-------------------------------------------------------------------------
	
	//prepare query plans, to avoid the overhead of finding query plans at
	//query execution times
	void preparePlan(string plan);
private:
	
	PGconn* conn;	//PostgreSQL Connection
	
	//store the edges in vectors from the database results
	void storeNodes(
			PGresult* res, //database resource
			vector<int>& graphs,//return value
			vector<int>& nodes, //return value
			vector<Attributes*>& vattrs	//return value
			);
	
	//store the edges in vectors from the database results
	void storeEdges(
			PGresult* res, //database resource
			vector<int>& graphs, //return value
			vector<int>& node1s, //return value
			vector<int>& node2s, //return
			vector<Attributes*>& vattrs	//return value
			);
	
	//-------------------------------------------------------------------------
	//	The wrapper for connecting to a PostgreSQL database
	//	conninfo: the connection information in the format of, e.g. 
	//			  "dbname = postgres user = postgres"
	//-------------------------------------------------------------------------
	inline void  connectDB(char* conninfo)
	{
		conn=PQconnectdb(conninfo);
		
		//error handling
		if(PQstatus(conn)!=CONNECTION_OK)
		{
			std::cerr<<"Connection to database "<<PQdb(conn)<<" failed"<<endl;
			std::cerr<<PQerrorMessage(conn)<<endl;
			exit_nicely();
		}
	}
	
	//-------------------------------------------------------------------------
	//	The wrapper for disconnecting to a PostgreSQL database
	//-------------------------------------------------------------------------
	inline void disconnectDB()
	{
		PQfinish(conn);
	}

	//-------------------------------------------------------------------------
	//	The wrapper for exiting a PostgreSQL database when exception happens
	//-------------------------------------------------------------------------
	inline void exit_nicely()
	{
		PQfinish(conn);
		exit(1);
	}

	//-------------------------------------------------------------------------
	//	The wrapper for issuing a non-query command to a PostgreSQL database
	//	cmd: the acutal sql statements, e.g. "create table temp(id int);"
	//-------------------------------------------------------------------------
	inline void  nonQueryCmd(const char* cmd)
	{
		//execute the sql statement
		PGresult* res;
		res=PQexec(conn, cmd);
		
		//error handling
		if (PQresultStatus(res) != PGRES_COMMAND_OK)
		{
			std::cerr<<"error occurs when execute "<<cmd<<endl;
			fprintf(stderr, "command failed: %s", PQerrorMessage(conn));
			PQclear(res);
			exit_nicely();
		}
		
		//clear resource allocated
		PQclear(res);
	}

	//-------------------------------------------------------------------------
	//	The wrapper for issuing a query command to a PostgreSQL database
	//	cmd: the acutal sql statements, e.g. "select count(*) from temp;"
	//  Output: resource for accessing the returned query results
	//  NOTE: the return resource has to be cleared outside this function using
	//		  PQclear(PGresult*)
	//-------------------------------------------------------------------------
	PGresult* queryCmd( const char* cmd)
	{
		//execute the sql statement
		PGresult* res;
		res=PQexecParams(conn, cmd, 0, NULL, NULL, NULL, NULL, 1);
		
		//error handling
		if (PQresultStatus(res) != PGRES_TUPLES_OK)
		{
			std::cerr<<"error occurs when execute "<<cmd<<endl;
			fprintf(stderr, "command failed: %s", PQerrorMessage(conn));
			PQclear(res);
			exit_nicely();
		}
		
		//return the resource for accessing the returned query results
		return res;
	}
};

/*
 *   create table node(
 * 						graphID int4, 
 * 						nodeID int4, //note that for SAGA, the nodeID<2^16
 * 						attrName varchar, 
 * 						attrType char(1), 
 * 						attrValue varchar
 * 						);
 */
struct NodeRec
{
	uint32_t graphID;
	uint32_t nodeID;
	string attrName;
	char attrType;
	string attrValue;
};

/*
 *   create table edge(
 * 						graphID int4, 
 * 						node1ID int4, 
 * 						node2ID int4, 
 * 						attrName varchar, 
 * 						attrType char(1), 
 * 						attrValue varchar
 * 						);
 */

struct EdgeRec
{
	uint32_t graphID;
	uint32_t node1ID;
	uint32_t node2ID;
	string attrName;
	char attrType;
	string attrValue;
};

/*
 *   create table ortholog(
 * 						orthologID int4, 
 * 						orthologName varchar, 
 * 						source varchar
 * 						);
 *
 */ 

struct OrthologRec
{
	uint32_t orthologID;
	string orthologName;
	string source;
};

/*
 *   create table distance(
 * 						graphID int4, 
 * 						node1ID int4, 
 * 						node2ID int4, 
 * 						distance int4
 * 						);
 */

struct DistanceRec
{
	uint32_t graphID;
	uint32_t node1ID;
	uint32_t node2ID;
	uint32_t distance;
};

/*
 *   create table neighborhood(
 * 						orthlogyID int2 , 
 * 						degree int4, 
 * 						nbConnection int4, 
 * 						nbArray bit(96), 
 * 						graphID int4, 
 * 						nodeID int4
 * 						);
 */

struct NeighborhoodRec
{
	uint32_t orthologID;
	uint32_t degree;
	uint32_t nbConnection;
	uint16_t nbArray[96/16];
	uint32_t graphID;
	uint32_t nodeID;
};

/*
 *   create table fragment(
 * 						orthSeq int8, 
 * 						distSum int4, 
 * 						graphID int4, 
 * 						nidSeq int4[], 
 * 						distMatrix int4[]
 * 						);
 */

struct FragmentRec
{
	uint64_t orthSeq;
	uint32_t distSum;
	uint32_t graphID;
	vector<uint32_t> nidSeq;
	vector<uint32_t> distMatrix;
};

class WriteToFile
{
public:
	WriteToFile(FILE* nf, FILE* ef, FILE* of, FILE* df, FILE* ff, FILE* nbf)
	{
		nodeFile=nf;
		edgeFile=ef;
		orthFile=of;
		distFile=df;
		neighborFile=nbf;
		fragmentFile=ff;
	}
	void writeNodeRec(NodeRec rec);
	void writeEdgeRec(EdgeRec rec);
	void writeOrthologRec(OrthologRec rec);
	void writeDistanceRec(DistanceRec rec);
	void writeNeighborhoodRec(NeighborhoodRec rec);
	void writeFragmentRec(FragmentRec rec);
private:
	FILE* nodeFile;
	FILE* edgeFile;
	FILE* orthFile;
	FILE* distFile;
	FILE* neighborFile;
	FILE* fragmentFile;
};

void getIntVector(char* ptr, vector<int>& vec);

int difference(const vector<int>& vec1, const vector<int>& vec2);

#endif
