/******************************************************************************
 * Definition for some utility functions for manipulating graphs
 * Author: Yuanyuan Tian (ytian@umich.edu)
 * ***************************************************************************/
#include "graphio.h"
#include <hash_map.h>
#include "../util/util.h"
HashGraph* readGraphSAGAFormat(
		FILE* f,	//input file in SAGA graph format
		Ortholog& orth,	//ortholog name to ortholog id mapping
		OrthologInfoList& orthinfolist	//return value, the orthology info
		)
{
	orthinfolist.clear();
	HashGraph* retG=new HashGraph();
	char line[1500];
	string name;
	//int graphCount=0;
	int ret;

	//read graph name
	ret=fscanf(f, "%s\n", line);

	if(ret==EOF) return NULL;
	//debug(2, line<<endl);
	
	name=string(line);
	//name=trimStr(name); 
	//cout<<"after trim: "<<name<<endl;			

	//read nodes
	int numNode;

	ret=fscanf(f, "%d\n", &numNode);
	assert(ret!=EOF);
	
	AttributeValue attrval;
	attrval.type='s';
	attrval.value=name;
	
	//set graph attributes
	Attributes* gattrs=new Attributes;
	gattrs->insert(Attributes::value_type("name", attrval));
	retG->setGraphAttrs(gattrs);

	for(int i=0;i<numNode;i++){	//read node
		int numFmly;
		ret=fscanf(f, "%s\t%d", line, &numFmly);
		assert(ret!=EOF);
		string nodelabel=string(line);
		if(nodelabel.size()>48)
			nodelabel=nodelabel.substr(0, 48)+"..";
		
		Attributes* nattrs=new Attributes;
		attrval.type='s';
		attrval.value=nodelabel;
		nattrs->insert(Attributes::value_type("name", attrval));
		OrthologInfo orthinfo;
		char buff[50];
		for(int j=0; j<numFmly; j++)
		{
			ret=fscanf(f, "\t%s", line);
			assert(ret!=EOF);
			uint32_t orthid=orth.getOrthologID(string(line));
			orthinfo.push_back(orthid);
			attrval.type='i';
			sprintf(buff, "%u", orthid);
			attrval.value=string(buff);
			nattrs->insert(Attributes::value_type("orthology", attrval));
		}
		retG->addNode(nattrs);
		orthinfolist.push_back(orthinfo);
	}
	//read edges
	int numEdge;
	ret=fscanf(f, "%d\n", &numEdge);
	assert(ret!=EOF);

	int node1, node2;
	float weight;
	for(int j=0; j<numEdge; j++)
	{
		ret=fscanf(f, "%d\t%d\t%f\n",  &node1, &node2, &weight);
		assert(ret!=EOF);

		if(node1!=node2)
		{
			Attributes* eattrs=new Attributes;
			attrval.type='b';
			attrval.value="false";
			eattrs->insert(Attributes::value_type("directed", attrval));
			retG->setEdge(node1, node2, eattrs);
		}	
	}
	
	return retG;
}

HashGraph* createGraphFromDB(
		DB* db,	//the database from where the graph is read from
		string tablePrefix,	//the prefix of tables in the database
		int graphID,	//the graphID of the graph in the database	
		string nodeCondition, //the condition on node table
		string edgeCondition, //the condition on edge table
		OrthologInfoList& orthinfolist  //return value, the orthology info
		)
{
	orthinfolist.clear();
	debug(2, "Read Graph "<<graphID<<endl);
	HashGraph* retG=new HashGraph();
	Attributes* pattrs=new Attributes;
	db->getGraphAttr(tablePrefix+"_graph", graphID, *pattrs, "");
	AttributeValue av;
	av.type='s';
	av.value=nodeCondition;
	
	if(nodeCondition!="")
		pattrs->insert(Attributes::value_type("nodeCondition", av));
	
	av.value=edgeCondition;
	if(edgeCondition!="")
		pattrs->insert(Attributes::value_type("edgeCondition", av));
	
	retG->setGraphAttrs(pattrs);
	
	char buff[1000];
	if(nodeCondition=="")
	{
		sprintf(buff, "where graphID=%d", 
			tablePrefix.c_str(), graphID);
	}else
	{
		sprintf(buff, "where (graphid, nodeid) in (select graphid, nodeid from "
				"%s_node where graphID=%d and %s)", tablePrefix.c_str(), graphID, 
				nodeCondition.c_str());
	}
	string condition=string(buff);
	vector<int> graphs;
	vector<int> node1s, node2s;
	vector<Attributes*> vattrs;
	db->selectNodes(tablePrefix+"_node", condition, graphs, node1s, vattrs);
	
	hash_map<int, int, inthash> nodemap;
	
	for(unsigned int i=0; i<vattrs.size(); i++)
	{
		nodemap[node1s[i]]=i;
		debug(2, "add node "<<node1s[i]<<endl);
		retG->addNode(vattrs[i]);
		pair<Attributes::const_iterator, Attributes::const_iterator> p =
			vattrs[i]->equal_range("orthology");
		OrthologInfo orthinfo;
		for (Attributes::const_iterator j = p.first; j != p.second; ++j)
			orthinfo.push_back(atoi(j->second.value.c_str()));
		orthinfolist.push_back(orthinfo);
	}
	
	graphs.clear();
	node1s.clear();
	vattrs.clear();
	
	if(edgeCondition=="")
	{
		sprintf(buff, "where graphID=%d", 
			tablePrefix.c_str(), graphID);
	}else
	{
		sprintf(buff, "where (graphid, node1id, node2id) in (select graphid, "
				"node1id, node2id from %s_edge where graphID=%d and %s)", 
			tablePrefix.c_str(), graphID, edgeCondition.c_str());
	}
	condition=string(buff);
	
	db->selectEdges(tablePrefix+"_edge", condition, graphs, node1s, node2s, vattrs);
	
	for(unsigned int i=0; i<vattrs.size(); i++)
	{
		retG->setEdge(nodemap[node1s[i]], nodemap[node2s[i]], vattrs[i]);
	}
	
	return retG;
}

HashGraph* readGraphFromDB(
		DB* db,	//the database from where the graph is read from
		string tablePrefix,	//the prefix of tables in the database
		int graphID,	//the graphID of the graph in the database	
		OrthologInfoList& orthinfolist  //return value, the orthology info
		)
{
	orthinfolist.clear();
	debug(2, "Read Graph "<<graphID<<endl);
	HashGraph* retG=new HashGraph();
	Attributes* pattrs=new Attributes;
	db->getGraphAttr(tablePrefix+"_graph", graphID, *pattrs, "");
	retG->setGraphAttrs(pattrs);
	
	char buff[50];
	sprintf(buff, "where graphID=%d", graphID);
	string condition=string(buff);
	vector<int> graphs;
	vector<int> node1s, node2s;
	vector<Attributes*> vattrs;
	db->selectNodes(tablePrefix+"_node", condition, graphs, node1s, vattrs);
	
	debug(2, "graphs.size()="<<graphs.size()<<endl);
	debug(2, "node1s.size()="<<node1s.size()<<endl);
	debug(2, "vattrs.size()="<<vattrs.size()<<endl);
	
	for(unsigned int i=0; i<vattrs.size(); i++)
	{
		debug(2, "add node "<<node1s[i]<<endl);
		retG->addNode(vattrs[i]);
		pair<Attributes::const_iterator, Attributes::const_iterator> p =
			vattrs[i]->equal_range("orthology");
		OrthologInfo orthinfo;
		for (Attributes::const_iterator j = p.first; j != p.second; ++j)
			orthinfo.push_back(atoi(j->second.value.c_str()));
		orthinfolist.push_back(orthinfo);
	}
	
	graphs.clear();
	node1s.clear();
	vattrs.clear();
	
	db->selectEdges(tablePrefix+"_edge", condition, graphs, node1s, node2s, vattrs);
	debug(2, "graphs.size()="<<graphs.size()<<endl);
	debug(2, "node1s.size()="<<node1s.size()<<endl);
	debug(2, "node2s.size()="<<node2s.size()<<endl);
	debug(2, "vattrs.size()="<<vattrs.size()<<endl);
	for(unsigned int i=0; i<vattrs.size(); i++)
	{
		retG->setEdge(node1s[i], node2s[i], vattrs[i]);
		debug(2, "add edge ("<<node1s[i]<<","<<node2s[i]<<")"<<endl);
	}
	
	return retG;
}

void writeDGFAttrbutesHeader(ostream& s, Attributes* pattrs, map<string, vector<string> >& attrMap)
{
	Attributes::const_iterator iter;
	for(iter=pattrs->begin(); iter!=pattrs->end(); iter++)
	{
		if(iter->first=="name" || iter->first=="orthology")
			continue;
		if(attrMap.find(iter->first)!=attrMap.end())
			continue;
		writeGDFAttribute(s, iter->first, iter->second.type);
		vector<string> temp;
		attrMap.insert(map<string, vector<string> >::value_type(iter->first, temp));
	}
	s<<endl;
}

void readAttributeValues(Attributes* pattrs, map<string, vector<string> >& attrMap)
{
	map<string, vector<string> >::iterator pmap;
	for(pmap=attrMap.begin(); pmap!=attrMap.end(); pmap++)
		pmap->second.clear();
	Attributes::const_iterator iter;
	for(iter=pattrs->begin(); iter!=pattrs->end(); iter++)
	{
		if(iter->first=="name" || iter->first=="orthology")
			continue;
		assert(attrMap.find(iter->first)!=attrMap.end());
		attrMap[iter->first].push_back(iter->second.value);
	}
}

void writeAttributeValues(ostream& s, map<string, vector<string> >& attrMap)
{
	map<string, vector<string> >::iterator pmap;
	for(pmap=attrMap.begin(); pmap!=attrMap.end(); pmap++)
	{
		string str=toString(pmap->second, ':');
		replaceBadChar(str, ",", '|');
		s<<","<<str;
	}
	s<<endl;
}

HashGraph* readGraphGDFFormat(
		FILE* f,	//input file in GDF format
		string gname, //the input graph name
		Ortholog& orth,	//ortholog name to ortholog id mapping
		OrthologInfoList& orthinfolist,	//return value, the orthology info
		bool directed
		)
{
	if(feof(f)) return NULL;
	orthinfolist.clear();
	HashGraph* retG=new HashGraph();
	AttributeValue attrval;
	attrval.type='s';
	attrval.value=gname;
	
	//set graph attributes
	Attributes* gattrs=new Attributes;
	gattrs->insert(Attributes::value_type("name", attrval));
	retG->setGraphAttrs(gattrs);
	debug(31, "set graph attribute\n");
	
	string str;
	bool readnode=true;
	int orthIndex=-1;
	vector<string> attrnames;
	vector<char> attrtypes;
	vector<string> defvalues;
	
	hash_map<string, int, strhash, eqstr> nodemap;
	
	string attrname, defvalue;
	char attrtype;
	char buff[50];
	while(readOneLine(f, str))
	{
		debug(31, str<<endl);
		//now parse the string
		if(str.substr(4, 4)=="def>")
		{
			debug(31, "readnode = "<<readnode<<endl);
			if(str.substr(0, 8)=="nodedef>")
				readnode=true;
			else
				readnode=false;
			debug(31, "readnode = "<<readnode<<endl);
			attrnames.clear();
			attrtypes.clear();
			defvalues.clear();
			int start=8;
			string seg;
			
			do{
				start=getOneSegment(str, start, ',', seg);
				debug(31, "start="<<start<<endl);
				debug(31, "string::npos="<<string::npos<<endl);
				parseGDFAttribute(seg, attrname, attrtype, defvalue);
				if(attrname=="orthologies")
					orthIndex=attrnames.size();
				attrnames.push_back(attrname);
				attrtypes.push_back(attrtype);
				defvalues.push_back(defvalue);
			}while(start!=string::npos);
	
		}else
		{
			debug(31, "read values"<<endl);
			debug(31, "orthIndex: "<<orthIndex<<endl);
			if(readnode)
			{
				assert(attrnames[0]=="name");
				int start=0;
				string seg;
				int i=0;
				Attributes* nattrs=new Attributes;
				OrthologInfo orthinfo;
				
				do{
					start=getOneSegment(str, start, ',', seg);
					if(i==0)
						nodemap[seg]=orthinfolist.size();
					debug(31, "getOneSegment: "<<seg<<endl);
					if(i==orthIndex)//if the attribute name is orthologies
					{
						debug(31, "orthologies\n");
						string orthlabel;
						int p=0;
						do{
							p=getOneSegment(seg, p, ':', orthlabel);
							debug(31, "orth label: "<<orthlabel);
							uint32_t orthid=orth.getOrthologID(orthlabel);
							orthinfo.push_back(orthid);
							attrval.type='i';
							sprintf(buff, "%u", orthid);
							attrval.value=string(buff);
							nattrs->insert(Attributes::value_type("orthology", attrval));
						}while(p!=string::npos);
					}else
					{
						attrval.type=attrtypes[i];
						if(seg.empty())
						{
							attrval.value=defvalues[i];
							nattrs->insert(Attributes::value_type(attrnames[i], attrval));
						}
						else
						{
							vector<string> items;
							readItems(seg, ':', items);
							for(unsigned int ii=0; ii<items.size(); ii++)
							{
								attrval.value=items[ii];
								nattrs->insert(Attributes::value_type(attrnames[i], attrval));
							}
						}
					}
					
					i++;
				}while(start!=string::npos);
				
				if(orthIndex<0)
				{
					attrval.type='i';
					attrval.value="0";
					nattrs->insert(Attributes::value_type("orthology", attrval));
					orthinfo.push_back(0);
				}
				
				retG->addNode(nattrs);
				orthinfolist.push_back(orthinfo);
				
			}else
			{
				assert(attrnames[0]=="node1" && attrnames[1]=="node2");
				int start=0;
				string seg;
				int i=0;
				Attributes* eattrs=new Attributes;
				int n1, n2;
				
				do{
					start=getOneSegment(str, start, ',', seg);
					if(i==0)
					{
						n1=nodemap[seg];
					}
					else if(i==1)
					{
						n2=nodemap[seg];
					}else
					{
						attrval.type=attrtypes[i];
						if(seg.empty())
						{
							attrval.value=defvalues[i];
							eattrs->insert(Attributes::value_type(attrnames[i], attrval));
						}
						else
						{
							vector<string> items;
							readItems(seg, ':', items);
							for(unsigned int ii=0; ii<items.size(); ii++)
							{
								attrval.value=items[ii];
								eattrs->insert(Attributes::value_type(attrnames[i], attrval));
							}
						}
					}
					i++;
				}while(start!=string::npos);
				if (directed==false)
					retG->setEdge(n1, n2, eattrs);
				else
					retG->setDirectedEdge(n1, n2, eattrs);
			}
		}
	}
	return retG;
}



//assume that all the nodes have the same attributes
//assume that all the edges have the same attributes
void writeGraphGDFFormat(
		ostream& s,	//output in GDF format
		HashGraph* G, //input graph
		Ortholog& orth,	//ortholog name to ortholog id mapping
		OrthologInfoList& orthinfolist	//the orthology info
		)
{
	if(G->n()==0) return;
	
	//write the header for nodes
	s<<"nodedef>name,orthologies varchar";
	map<string, vector<string> > attrMap;
	Attributes* pattrs=G->getNodeAttrs(0);
	writeDGFAttrbutesHeader(s, pattrs, attrMap);

	//write nodes
	for(int v=0; v<G->n(); v++)
	{
		pattrs=G->getNodeAttrs(v);	
		//write name
		string name=pattrs->find("name")->second.value;
		replaceBadChar(name, ",", '|');
		s<<name<<",";
		//write orthology
		vector<string> orths;
		for(unsigned int i=0; i<orthinfolist[v].size(); i++)
			orths.push_back(orth.getOrthologName(orthinfolist[v][i]));
		string orthstr=toString(orths, ':');
		replaceBadChar(orthstr, ",", '|');
		s<<orthstr;
		
		readAttributeValues(pattrs, attrMap);
		writeAttributeValues(s, attrMap);
	}
	
	//write the header for edges & write edges
	attrMap.clear();
	int numEdge=-1;
	for(int v=0; v<G->n(); v++)
	{
		EdgeMap* p_neighbors=G->getNeighbors(v);
		EdgeMap::iterator pnb;
		for (pnb=p_neighbors->begin(); pnb!=p_neighbors->end(); pnb++)
		{
			int w=pnb->first;
			if(v>w)
				continue;
			numEdge++;
			if(numEdge==0)
			{
				//write header
				s<<"edgedef>node1,node2";
				pattrs=G->getEdgeAttrs(v, w);
				writeDGFAttrbutesHeader(s, pattrs, attrMap);
			}
			
			//write the edges
			string node1=G->getNodeAttrs(v)->find("name")->second.value;
			string node2=G->getNodeAttrs(w)->find("name")->second.value;
			replaceBadChar(node1, ",", '|');
			replaceBadChar(node2, ",", '|');
			s<<node1<<","<<node2;
			readAttributeValues(pattrs, attrMap);
			writeAttributeValues(s, attrMap);
		}
	}
}
