/******************************************************************************
 * The definition of class LoadGraph which loads the graph into database. 
 * Author: Yuanyuan Tian (ytian@umich.edu)
 * ***************************************************************************/

#include "load.h"
#include "../debug/debug.h"

void LoadGraph::insertFragmentTuple(vector<NodeOrtholog>& vfset)
{
	vector<NodeOrtholog> temp(vfset);
	std::sort(temp.begin(), temp.end(), ltNO);

	vector<int> orths;
	
	FragmentRec rec;
	for(unsigned int i=0; i<temp.size(); i++)
	{
		rec.nidSeq.push_back(temp[i].nodeID);
		orths.push_back(temp[i].orthID);
	}
	
	for(unsigned int i=0; i<temp.size()-1; i++)
		for(unsigned int j=i+1; j<temp.size(); j++)
			rec.distMatrix.push_back(D[temp[i].nodeID][temp[j].nodeID]);
	
	rec.orthSeq=getOrthSeq(orths);
	string tempstr;
	getDistSeqSum(rec.distMatrix, rec.distSum, tempstr);
	rec.graphID=gid;
	wtf->writeFragmentRec(rec);
}

//a recursive function, helper function for insertFragment
void LoadGraph::insertFragmentHelp(NodeSet & vs, 
	unsigned int iterIndx, vector<NodeOrtholog>& vfset)
{
	if(iterIndx<vs.size())
	{
		for(unsigned int i=0; i<(*pOrthinfolist)[vs[iterIndx]].size(); i++)
		{
			NodeOrtholog vf;
			vf.orthID=(*pOrthinfolist)[vs[iterIndx]][i];
			vf.nodeID=vs[iterIndx];
			vfset.push_back(vf);
			
			insertFragmentHelp( vs,iterIndx+1, vfset);
			vfset.pop_back();
		}
		
	}else
		insertFragmentTuple( vfset);
}

//enumerate all the possible vertex set upto size fragSize+1, and insert into the fragment table of the database
//fragSize+1 is the fragSize+1 value (see paper), gid is the gid in the Graph table of the database
//D is the distance matrix
// generate closely connected vertex set
//can reduce the size of the table significantly 
void LoadGraph::insertFragments()
{
	assert(fragSize>=2);
	NodeSetList list1, list2;

	for(int v1=0; v1<G->n()-fragSize+1; v1++)
	{
		for(int v2=v1+1; v2<G->n()-fragSize+2; v2++)
		{
			if((*pOrthinfolist)[v1][0]==0 || (*pOrthinfolist)[v2][0]==0)
				continue;
			NodeSet vs;
			vs.push_back(v1);
			vs.push_back(v2);

			if(fragSize==2 && D[v1][v2]<=fragLimit)
			{
				vector<NodeOrtholog> vfset;
				insertFragmentHelp(vs, 0, vfset);
			}
			else
				list1.push_back(vs);
		}
	}

	NodeSetList* prev_list, *curr_list;
	prev_list=&list1; curr_list=&list2;

	//we have k-subset, we want to generat k+1-subset
	for(int k=2; k<fragSize; k++)
	{		
		unsigned int numloop=prev_list->size();
		
		NodeSetList::iterator prev_it;
		
		for(unsigned int count=0; count<numloop; count++)
		{
			//debug(15, "count= "<<count<<endl);
			prev_it=prev_list->begin();
			
			NodeSet pvs=(*prev_it);//the vertex set

			int i=pvs[pvs.size()-1]+1; //vertex set sorted by id
			for(; i<G->n()-fragSize+k+1; i++)
			{			
				//debug(15, "i="<<i<<endl);
				if((*pOrthinfolist)[i][0]==0)
					continue;
			
				NodeSet vs(pvs);
				vs.push_back(i);
				
				if((int)vs.size()==fragSize)
				{
					//only work for fragment size=3
					int numSmall=0;
					bool ok=false;
					for(unsigned int s=0; s<vs.size()-1; s++)
					{
						for(unsigned int t=s+1; t<vs.size(); t++)
						{
							if(D[ vs[s] ][ vs[t] ]<=fragLimit)
							{
								numSmall++;
								if(numSmall>=2)
								{ ok=true; break;}		
							}							
						}
						if(ok) break;
					}
					if(ok)
					{
						vector<NodeOrtholog> vfset;
						insertFragmentHelp(vs, 0, vfset);
					}
				}
				else
					curr_list->push_back(vs);
				//debug(15, "after something");
			}
			
			prev_list->erase(prev_it);
		}
		//swap curr_map and prev_map
		NodeSetList* temp=prev_list;
		prev_list=curr_list;
		curr_list=temp;
	}	
}

//given a graph, distance matrix and confidence matrix, insert it into the table Graph
void LoadGraph::insertGraph()
{
	//insert G into the Graph Table
	db->insertIntoGraphTable(tablePrefix+"_graph", gid, *(G->getGraphAttrs()));
}

void LoadGraph::insertNodes()
{
	//insert nodes to the Node Table
	NodeRec nr;
	nr.graphID=gid;
	for(int i=0; i<G->n(); i++)
	{
		nr.nodeID=i;
		Attributes* p=G->getNodeAttrs(i);
		
		Attributes::iterator iter;
		for(iter=p->begin(); iter!=p->end(); iter++)
		{
			nr.attrName=iter->first;
			nr.attrType=iter->second.type;
			nr.attrValue=iter->second.value;
			wtf->writeNodeRec(nr);
		}
	}
}

void LoadGraph::insertEdges()
{
	//insert edges into the Edge Table
	EdgeRec er;
	er.graphID=gid;
	EdgeMap::iterator mp;
	for(int i=0; i<G->n(); i++)
	{
		EdgeMap* m=G->getNeighbors(i);

		for(mp=m->begin(); mp!=m->end(); mp++)
		{
			if (directional==0)
			{
				if(i>mp->first)
				continue;
			}
			//printf("edges\n");
			er.node1ID=i;
			er.node2ID=mp->first;
			//printf("node1:%d\t", er.node1ID);
			//printf("node2:%d\n", er.node2ID);
			//getchar();
			Attributes* p=mp->second.p_eattr;
			Attributes::iterator iter;
			for(iter=p->begin(); iter!=p->end(); iter++)
			{
				er.attrName=iter->first;
				er.attrType=iter->second.type;
				er.attrValue=iter->second.value;
				wtf->writeEdgeRec(er);
			}
		}
	}
}

void LoadGraph::insertOrthologs()
{
	//insert edges into the Edge Table
	OrthologRec otr;
	otr.source="unknown";
	OrthologID2NameMap* pmap=pOrth->getID2NameMap();
	OrthologID2NameMap::iterator iter=pmap->begin();
	for(;iter!=pmap->end(); iter++)
	{
		otr.orthologID=iter->first;
		otr.orthologName=iter->second;
		wtf->writeOrthologRec(otr);
	}
}

void LoadGraph::insertDistances()
{
	//insert pairwise distances into Distance table
	//assume undirected graph
	DistanceRec dr;
	dr.graphID=gid;
	for(int v1=0; v1<G->n()-1; v1++)
		for(int v2=v1+1; v2<G->n(); v2++)
		{
			if(D[v1][v2]>=INFIN)
				continue;
			dr.node1ID=v1;
			dr.node2ID=v2;
			dr.distance=D[v1][v2];
			debug(1, "in insertDistance\n");
			debug(1, "distance= "<<dr.distance<<endl);
			wtf->writeDistanceRec(dr);
		}
}

void LoadGraph::insertNeighborhoods()
{	
	NeighborhoodRec r;
	r.graphID=gid;
	
	EdgeMap::iterator mp, mp2;

	for(int v1=0; v1<G->n(); v1++)
	{
		r.degree=G->degree(v1);
		r.nbConnection=0;
		r.nodeID=v1;
		memset((char*)r.nbArray, 0, sizeof(uint16_t)*96/16);
		debug(1, "after memset\n");

		EdgeMap* m=G->getNeighbors(v1);
		for(mp=m->begin(); mp!=m->end(); mp++)
		{
			UINT v2=mp->first;
			
			for(unsigned int i=0; i<(*pOrthinfolist)[v2].size(); i++)
			{
				unsigned int pos=hashfunc((*pOrthinfolist)[v2][i]) % 96;
				setBit(r.nbArray, pos);
				debug(1, "set bit: "<<pos<<endl);
			}
			mp2=mp;
			mp2++;
			for(; mp2!=m->end(); mp2++)
			{
				UINT v3=mp2->first;
				if(G->isEdge(v2,v3))
					r.nbConnection++;
				debug(1, "checked edge ("<<v2<<", "<<v3<<")\n");
			}
		}
		
		for(unsigned int i=0; i<(*pOrthinfolist)[v1].size(); i++)
		{
			r.orthologID=(*pOrthinfolist)[v1][i];
			debug(1, "before writeNeighborhoodRec\n");
			wtf->writeNeighborhoodRec(r);
			debug(1, "after writeNeighborhoodRec\n");
		}
	}
}

void LoadGraph::load(bool forsaga, bool fortale)
{
	//basic indices
	//insert into Graph, Vertex and Edge tables (files on disk)
	insertGraph();
	debug(33, "insert into graph succeeds\n");
	insertNodes();
	debug(33, "insert into node succeeds\n");
	insertEdges();
	debug(33, "insert into edge succeeds\n");
//	insertOrthologs();

	if(forsaga && G->e()>0)
	{
		D=new int* [G->n()];
		for(int i=0; i<G->n(); i++)
		{
			D[i]=new int[G->n()];
		}
		//get distance matrix
		floyd_Min_Dist(G, D);
		debug(33, "got distance matrix!\n");
		
		insertDistances();
		debug(33, "insert into distance succeeds\n");
		//enumerate vertex subset and insert into fragment Table
		insertFragments();
		//deallocate 
		for(int i=0; i<G->n(); i++)
		{
			delete [] D[i];
		}
		delete [] D;
		//debug(33, "insert into fragments succeeds\n");
	}
	
	if(fortale)
	{
		//enumerate neighborhoods and insert into the Neighborhood table (file on disk)
		insertNeighborhoods();
		debug(33, "insert into neighborhoods succeeds\n");
	}
}
