/******************************************************************************
 * Definition for some utility functions for manipulating graphs
 * Author: Yuanyuan Tian (ytian@umich.edu)
 * ***************************************************************************/
#include "graphutil.h"
#include "minheap.h"
#include "util.h"

//Get the adjacency matrix of the graph
void getMatrices(
		HashGraph* G,	//input graph
		int** D			//output matrix (D is allocated outside)
		)
{
	for(int i=0; i<G->n(); i++)
	{
		for(int j=i+1; j<G->n(); j++)
		{
			if(G->isEdge(i,j))
				D[i][j]=D[j][i]=1;
			else
				D[i][j]=D[j][i]=INFIN;
		}
		D[i][i]=0;
	}
}

//Compute all pairwise shortest distance using Floyd's algorithm
//NOTE: The distance does not consider the weight of edge.
void floyd_Min_Dist(
		HashGraph* G,	//input graph 
		int** D			//output distance matrix
						//(D is allocated outside)
		) 
{  	
	for (int i=0; i<G->n(); i++) // Initialize D and Conf
		for (int j=0; j<G->n(); j++)
		{
			if(i==j)
			{
				D[i][j]=0;
			}
			else if(!G->isEdge(i, j))
			{
				D[i][j]=INFIN;
			}
			else
			{
			 	D[i][j]= 1;
			}
		}
	
	for (int k=0; k<G->n(); k++) // Compute all k paths
		for (int i=0; i<G->n(); i++)
		  for (int j=0; j<G->n(); j++)
		  {
			    if (i!=j && D[i][j]> (D[i][k]+ D[k][j]) )
			    {
			    		D[i][j]= D[i][k]+ D[k][j];
			    }
		  }
}

//Find the nodes on the shortest path between v1 and v2
//NOTE: The distance does not consider the weight of edge.
bool getBridge(
		HashGraph* G,	//input graph 
		int v1, 		//one end node
		int v2, 		//the other end node
		int len, 		//length of the shortest path (input)
		set<int>& vset	//returned set of ndoes
		) 
{ 
	// Depth first search
	vector<bool> mark(G->n(), false);
	mark[v1]=true;
	bool ret=false;
	//ending condition
	if( 1==len )
	{
		if( G->isEdge(v1, v2) )
		{
			vset.insert(v2);
			ret= true;
		}
		else
			ret= false;
	}
	else
	{
		EdgeMap* p_neighbors=G->getNeighbors(v1);
		EdgeMap::iterator pnb;
        for (pnb=p_neighbors->begin(); pnb!=p_neighbors->end(); pnb++)
        {
                UINT w=pnb->first;
                if ( !mark[w] && getBridge(G, w, v2, len-1, vset))
    			{
    				ret=true;
    				break;
    			}
        }
	}

	//back track
	mark[v1]=false;

	if(ret)
		vset.insert(v1);
	
	return ret;
}

//Compute the shortest paths from s to other nodes in the graph
//Dijkstra's shortest paths algorithm w/ priority queue
//NOTE: The distance does not consider the weight of edge.
void Dijkstra(
		HashGraph* G,		//input graph 
		int* D, 		//returned distance matrix
		int s			//source node
		) 
{
	int i, v, w;            // v is current vertex
	for(i=0; i<G->n(); i++)
	D[i]=INFIN;
	D[s]=0;
	// Heap array with lots of space
	vector<ELEM<int, int> > heapArray(G->e()) ;
	//Initialize heap array
	heapArray[0].key=0;
	heapArray[0].value=s;  
	
	// Create heap
	MinHeap<int, int, inthash, equal_to<int> > H(&heapArray, 1, G->e());
	vector<bool> mark(G->n(), false);
	
	ELEM<int, int> temp;
	
	for (i=0; i<G->n(); i++) 
	{   // Now, get distances
		do 
		{
			if(H.heapSize()<=0) return; // Nothing to remove
			temp=H.removeMin();
			v = temp.value;
		} while (mark[v]);
		
		mark[v]=true;
		
		if (D[v] == INFIN) return;  // Unreachable vertices
		EdgeMap* p_neighbors=G->getNeighbors(v);
		EdgeMap::iterator pnb;
        for (pnb=p_neighbors->begin(); pnb!=p_neighbors->end(); pnb++)
        {
            w=pnb->first;
            if (D[w] > (D[v] + 1)) 
			{ // Update D
				D[w] = D[v] + 1;
				temp.key = D[w]; temp.value = w;
				H.insert(temp);    // Insert new distance in heap
			}
        }
	}
}

//Compute the pairwise shortest distance among the set nodes in V
void getDistance(
		HashGraph* G, 	//input graph
		int** D,		//the returned distance matrix 
		hash_set<int>& V //The input set of nodes
		)
{
        int* dis=new int[G->n()];
        hash_set<int>::iterator p1, p2;
        for(p1=V.begin(); p1!=V.end(); p1++)
        {
                p2=p1;
                p2++;
                if(p2==V.end())
                        break;
                Dijkstra(G, dis, *p1);
                for(; p2!=V.end(); p2++)
                        D[*p1][*p2]=D[*p2][*p1]=dis[*p2];
        }
        delete [] dis;
}

//return the neighborhood of a node
void getNeighborhood(
		HashGraph* G, 	//input graph
		OrthologInfoList& orthinfolist,	//the orthology info for the ndoes in G 
		int v, 	//the input node
		Neighborhood& nb	//the output neighborhood
		)
{
	nb.v=v;
	nb.nbConn=0;
	vector<int> nodes;
	memset(nb.nbArray, 0, 96*sizeof(bool));
	EdgeMap* p_neighbors=G->getNeighbors(v);
	EdgeMap::iterator pnb;
	nb.degree=p_neighbors->size();
	if(nb.degree==0)
		return;

	for (pnb=p_neighbors->begin(); pnb!=p_neighbors->end(); pnb++)
	{
		int w=pnb->first;
		nodes.push_back(w);
		for(unsigned int i=0; i<orthinfolist[w].size(); i++)
		{
			unsigned int pos=hashfunc(orthinfolist[w][i]) % 96;
			nb.nbArray[pos]=true;
		}
	}
	   
	for(unsigned int i=0; i<nodes.size()-1; i++)
		for(unsigned int j=i+1; j<nodes.size(); j++)
		{
			if(G->isEdge(i,j))
				nb.nbConn++;
		}
}

void getAllNeighborhoods(
	HashGraph* G, 	//input graph
	OrthologInfoList& orthinfolist,	//the orthology info for the ndoes in G 
	vector<Neighborhood>& nbs	//the output neighborhoods
	)
{
	for (int i=0; i<G->n(); i++)
	{
		debug(46, "get neighborhoods for node "<<i<<endl);
		Neighborhood nb;
		getNeighborhood(G, orthinfolist, i, nb);
		nbs.push_back(nb);
	}
}

struct DegreePos {
        int degree;
        int pos;
};

bool gtDegree(const DegreePos d1, const DegreePos d2)
{
        return (d1.degree > d2.degree);
}

void topKImpNodes(
	HashGraph* G, 	//input graph
	OrthologInfoList& orthinfolist,  //the orthology info for the ndoes in G 
	int& k,	//num of important nodes
	vector<int>& nodes	//returned important nodes
	)
{
	vector<DegreePos> dpvec;
	for(int i=0; i<G->n(); i++)
	{
		if(orthinfolist[i].empty() || (orthinfolist[i].size()==1 && orthinfolist[i][0]==0))
			continue;
		DegreePos dp;
		dp.degree=G->degree(i);
		dp.pos=i;
		dpvec.push_back(dp);
	}
	if(k>(int)dpvec.size())
		k=dpvec.size();
	std::nth_element (dpvec.begin(), dpvec.begin()+(size_t)k, dpvec.end(), gtDegree);

	for(int i=0; i<k; i++)
		nodes.push_back(dpvec[i].pos);
}
